module mapper124 (CLK0, C_A, C_D, C_WR, P_A, M2, rst, h_blank, prgbank, chrbank, IRQ, mirror, auden, wramen, dipsw, coin_in, subdata);



	input wire CLK0; 				// system clock
	input wire [15:0] C_A;		// CPU add
	input wire [7:0] C_D;		// CPU data
	input wire C_WR;				// CPU WR
	input wire [13:0] P_A;		// PPU add
	input wire M2;					// CPU M2
	input wire rst;				// reset
	input wire h_blank;			// MMC3 PPU_A12 edge detector input
	input wire [7:0] dipsw;		// dipswitch. 0 = off, 1 = on
	input wire coin_in;			// 1 = coin inserted
	
	output reg [8:0] prgbank;		// 8K PRG bank 
	output reg [11:0] chrbank;		// 1K CHR bank
	output reg IRQ;					// IRQ 
	output reg [1:0] mirror;		// mirroring
	output reg [1:0] wramen;		// WRAM enable
	output wire [8:0] subdata = {((C_A[15:8] == 8'h4f) & &C_A[3:0]), ~coin_in, ~dipsw[2], ~dipsw[4], ~dipsw[6], ~dipsw[7], ~dipsw[5], ~dipsw[3], ~dipsw[1]};		// substitute data to CPU; bit 8 = enable substitute data when set.
	output reg auden;					// audio enable
	
	
	wire [5:0] mmc1prg;		// also has extra PRG banks for WRAM and 512K PRG stuff, not used here
	wire [4:0] mmc1chr;
	wire [1:0] mmc1mir;
	wire [1:0] mmc1wramen;	// not used here
	
	mmc1 thingera (	.CLK0(CLK0), .C_A(C_A), .C_D(C_D), .C_WR(C_WR), .P_A(P_A), .rst(rst),
							.active(1'b1), 
							.prgbank(mmc1prg), .chrbank(mmc1chr), .mirror(mmc1mir), 
							.wramen(mmc1wramen)
					  );

					  
	wire [5:0] mmc3prg;
	wire [7:0] mmc3chr;
	wire mmc3mir;
	wire mmc3irq;
	wire [1:0] mmc3wramen;
	
	mmc3 thingerb (	.CLK0(CLK0), .C_A(C_A), .C_D(C_D), .C_WR(C_WR), .P_A(P_A), .M2(M2), .rst(rst),
							.active(1'b1), .mmcb(1'b1),.h_blank(h_blank), 
							.prgbank(mmc3prg), .chrbank(mmc3chr), .IRQ(mmc3irq), .mirror(mmc3mir), 
							.wramen(mmc3wramen)
					  );
 
 
	// UNROM
	
	wire [2:0] unromprg = unrombank | {3{C_A[14]}};		// only 3 bits on this particular UNROM
	
	reg [2:0] unrombank;
	
	
	always @ (posedge CLK0)
	begin
		
		if (C_WR & C_A[15]) unrombank <= C_D[2:0];
		
	end
	
	
	
	// AMROM
	
	reg [3:0] amrombank;
	
	always @ (posedge CLK0)
	begin
	
		if (C_WR & C_A[15]) amrombank <= {C_D[4], C_D[2:0]};
		
	end
	
	
	// custom mapper
	
	reg [7:0] rega;	// 5001h reg
	reg [6:0] regb;	// 5002h reg
		
	always @ (posedge CLK0)
	begin
	
		if (rst)
		begin
			rega <= 0;
			regb <= 0;
			auden <= 0;
		end
		else if (C_WR)
		casex(C_A)
		16'b0101_0000_xxxx_0000	: auden <= C_D[0];
		16'b0101_0000_xxxx_0001	: rega <= C_D;
		16'b0101_0000_xxxx_0010	: regb <= C_D[6:0];
		endcase
				
	end
	
	// dat muxer logic
	
	// PRG muxing
	wire prga17 = (rega[5]) ? rega[0] : mmc3prg[4];		// select between reg bit or MMC3 PRG A17
	wire prga17b = (rega[5]) ? amrombank[2] : mmc3prg[4];	// select for AMROM
		
	always
	casex({((C_A[15:13] == 3'h3) & ~rega[5]), (rega[7] & C_A[15]), regb[5:4]})		// control EPROM en, mapper #
	4'b0100	: prgbank <= {rega[4:1], prga17 , unromprg[2:0], C_A[13]};		// select UNROM
	4'b0101	: prgbank <= {rega[4:1], prga17b, amrombank[1:0], C_A[14:13]};	// select AMROM
	4'b0110	: prgbank <= {rega[4:1], prga17 , mmc1prg[3:0]};				// select MMC1
	4'b0111	: prgbank <= {rega[4:1], prga17 , mmc3prg[3:0]};				// select MMC3
	4'b1xxx	: prgbank <= 9'h0;														// last bank = PRG WRAM
	default	: prgbank <= {6'h38, C_A[15:13]};									// select control EPROM if nothing else is selected at slot 7
	endcase
	
	// CHR muxing
	wire chra17 = (regb[6]) ? regb[0] : mmc3chr[7];		// select between reg bit or MMC3 CHR A17
	
	always
	casex({rega[6], regb[5:4]})
	3'b0xx	: chrbank <= {1'b1, 8'h0, P_A[12:10]};										// 8K VRAM
	3'b1x0	: chrbank <= {1'b0, regb[3:1], chra17, mmc1chr[4:0], P_A[11:10]};	// select CHR ROM for MMC1
	3'b1x1	: chrbank <= {1'b0, regb[3:1], chra17, mmc3chr[6:0]};					// select CHR ROM for MMC3
	endcase
	
	
	// IRQ muxing
	
	always
	case(regb[5:4])
	2'h0	: IRQ <= 1'b0;		// UNROM has no IRQ
	2'h1	: IRQ <= 1'b0;		// AMROM has no IRQ
	2'h2	: IRQ <= 1'b0;		// MMC1 has no IRQ
	2'h3	: IRQ <= mmc3irq;	// MMC3 does have IRQ
	endcase
	
	// WRAM muxing  
	
	always
	case(rega[5])
	1'h0	: wramen <= mmc3wramen;	// WRAM is controlled by MMC3
	1'h1	: wramen <= 2'b00;		// WRAM is off no matter what
	endcase
	
	// mirror muxing
	
	always
	case(regb[5:4])
	2'h0	: mirror <= 3'h2;								// UNROM mirror (fixed horizontal)
	2'h1	: mirror <= {2'b00,  amrombank[3]};		// AMROM mirror (mapper controlled)
	2'h2	: mirror <= {1'b0, mmc1mir};				// MMC1 mirror
	2'h3	: mirror <= {2'b01, mmc3mir};				// MMC3 mirror
	endcase
	
	
	
	
endmodule

